/*This file is part of a C version of SLINK a simulation program based
on the LINKAGE package. SLINK was originally described in: D. E.
Weeks, J. Ott, G. M. Lathrop (1990), SLINK: a general simulation
program for linkage analysis, American Journal of Human Genetics
47(1990), A204 (abstract). This version of SLINK is adapted to use the
faster pedigree traversal algorithms described in: R. W. Cottingham
Jr., R. M. Idury, A. A. Schaffer (1993), Faster sequential genetic
linkage computations, American Journal of Human Genetics 53(1993), pp.
252-263.*/
/* This file contains definitions sharec with the othe LINKAGE programs*/
/*March 14, 1994, A.A. Schaffer changed to make declarations more portable*/


#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#define STRICT_R_HEADERS
#include <R.h> //to use R memory management functions
/*include <p2c/p2c.h>*/

#define true            1
#define false           0
#define NULL            0
#define FileNotFound   10  /*From p2c.h*/

#ifndef EXIT_SUCCESS
#  ifdef vms
#define EXIT_SUCCESS 1
#define EXIT_FAILURE (02000000000L)
#else
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#  endif     /*From p2c.h*/
#endif
#ifndef Void
#define Void void
#endif

#ifndef Static
#define Static static
#endif

#ifndef Local
#define Local static
#endif

#define Free free
#define Malloc malloc

#define version         "2.51"   /*version of SLINK*/
/* SOME USER DEFINED CONSTANTS */
/*THE PROGRAM WILL TELL YOU IF THE FOLLOWING TWO CONSTANTS CAN
  BE REDUCED*/
/*IF THE PROGRAM TERMINATES WITH AN ERROR IN RECOMBINATION INCREASE MAXNEED*/

#define maxneed        157 /* 32 for 3-point, 157 for 4-point analysis */

/* The following introduced by Ramana */
#define maxclasssize    8  /* 8 for 4-point analysis, 16 for 5-point */
#define maxisozygclass  maxclasssize*maxclasssize 
#define maxprobclass    41 /* 41 for 4-point analysis, 122 for 5-point */

    /*MAXIMUM NUMBER OF RECOMBINATION PROBABILITIES*/
/*THE FOLLOWING SHOULD BE LARGER THAN MININT*/

#define maxcensor       50000L   /*MAXIMUM FOR CENSORING ARRAY*/
#define maxsys          40   /*MAXIMUM NUMBER OF SYSTEMS*/
#define maxlocus        8   /*MAXIMUM NUMBER OF LOCI */
#define maxrec          maxlocus   /*MAXIMUM POSSIBLE NUMBER OF RECOMB/MEI*/
#define maxseg          64   /*    (maxlocus-1)        */
/* = 2              I.E. 2 TO THE POWER maxlocus-1 */
#define maxall          11   /*MAX NUMBER OF ALLELES AT A SINGLE LOCUS*/
#define maxhap         180   /*MAXIMUM NUMBER OF HAPLOTYPES*/
/* = n1 x n2 x ... ETC. WHERE ni = NUMBER OF ALLELES LOCUS I*/
#define maxfem          maxhap * (maxhap + 1) / 2
    /*MAXIMUM NUMBER OF JOINT GENOTYPES FOR A FEMALE*/
/* = maxhap x (maxhap + 1 ) /2 */
#define maxmal          maxfem
    /*MAXIMUM NUMBER OF JOINT GENOTYPES FOR A MALE*/
/* = maxfem (AUTOSOMAL) OR maxhap (SEXLINKED)*/
#define maxind          5000   /*MAXIMUM NUMBER OF INDIVIDUALS*/
#define maxped          500   /*MAXIMUM NUMBER OF PEDIGREES*/
#define maxchild        16   /*MAXIMUM NUMBER OF FULLSIBS IN A SIBSHIP*/
#define maxloop         2   /*MAXIMUM NUMBER OF LOOPS PER PEDIGREE*/

#define minfreq         0.0

#define affall          2
/*DISEASE ALLELE FOR QUANTITATIVE TRAITS
                               OR AFFECTION STATUS*/
/* QUANTITATIVE TRAIT */
#define maxtrait        3
    /*MAXIMUM NUMBER OF QUANTITATIVE FACTORS AT A SINGLE LOCUS*/

#define missval         0.0   /*MISSING VALUES FOR QUANTITATIVE TRAITS */
/* AFFECTION STATUS */

#define missaff         0   /*MISSING VALUE FOR AFFECTION STATUS */
#define affval          2   /*CODE FOR AFFECTED INDIVIDUAL*/
#define maxliab         10   /*MAXIMUM NUMBER OF LIABILITY CLASSES */
/* BINARY (FACTOR UNION) SYSTEM */
#define maxfact         11
    /*MAXIMUM NUMBER OF BINARY CODES AT A SINGLE LOCUS*/
/* OTHERS */

#define scalemult       2.0   /*SCALE WEIGHT FOR EACH LOCUS INCLUDED*/

#define fitmodel        false
    /*TRUE IF ESTIMATING PARAMETERS OTHER THAN REC*/
#define dostream        true   /*STREAM FILE OUTPUT*/
#define byfamily        false   /*GIVE LOD SCORES BY FAMILY*/

#define zerolike        (-1.0e20)
    /*FOR INCONSISTENT DATA OR RECOMBINATION */
#define log10_          2.30259

#define minint          (-32767)   /*MINIMUM ALLOWED INTEGER*/
/*GRADIENT APPROXIMATIONS*/

#define approximate     false

/*next two typedefs come from p2c*/
typedef char boolean;
typedef short uchar;

typedef struct gennurec {
  long genenumber[maxhap][maxhap];
} gennurec;

typedef struct approxrec {
  uchar approxarray[maxped][maxfem];
} approxrec;

typedef struct censorrec {
  boolean censor[maxcensor - minint];
} censorrec;

typedef double genotype[maxfem];
typedef double covmatrix[maxtrait][maxtrait];
typedef long mutarray[3][2];
typedef double thesemeans[maxtrait];
typedef thesemeans means[maxall + 1][maxall];
typedef enum {
  auto_, mauto, sex, msex
} pathway;
typedef enum {
  peelup, peeldown
} direction;

typedef long binset;

typedef binset phenarray[maxall];


typedef enum {
  affection, quantitative, binary_, null_
} locustype;

/* typedef struct locusvalues {
  long nallele, format;
  double freq[maxall];
  struct locusvalues *privlocus;
  locustype which;
  union {
    struct {
      double pen[maxall + 1][maxall][3][maxliab];
      long nclass;
    } U0;
    struct {
      long ntrait;
      means pm;
      covmatrix vmat;
      double det, contrait, conmat;
    } U1;
    phenarray allele;
  } UU;
} locusvalues; */

typedef struct phenotype {
  locustype which;
  binset phenf;
  double x[maxtrait];
  boolean missing;
  long aff;
  long liability;
} phenotype;

typedef phenotype *hindphen[maxsys];
typedef phenotype *indphen[maxlocus];


/* a structure of type 'thisarray' stores the conditional
genotype probabilities for an individual. The
probabilities are stored in the array genarray.
We expect this array to be sparse. The field sparseflag
is a boolean array such that sparseflag[i] is nonzero
if and only if genarray[i] is nonzero. */

typedef struct thisarray {
  union {
    unsigned char sparseflag[maxfem];
  } sparsegen;
  genotype genarray;
} thisarray;

typedef boolean possvect[maxall][maxall];
typedef possvect possarray[maxlocus];

typedef struct information {
  possarray possible;
} information;

/* a record of type this person stores information about one person
   in one pedigree */

/*typedef struct thisperson {
  long id, ped, inloop;
  struct thisperson *pa, *ma, *foff, *nextpa, *nextma;
  thisarray *gen;
  hindphen holdphen;
  indphen phen;
  phenotype *privphen;
  boolean unknown, multi, done, up, male, firstpass;
  information *store;
} thisperson; */



typedef long subhap[maxseg];
typedef double thetarray[maxlocus];
typedef double happrob[maxneed];
unsigned nuneed; /* Introduced by Ramana, actual size of segprob arrays */

typedef struct thetavalues {
  thetarray theta,theta1, theta2;
  happrob segprob;
} thetavalues;


/*The following declarations help store a partial correspondence
between genotypes and haplotypes */
unsigned currentfence;           /* Ramana */
unsigned long base[maxfem], fence[maxfem];
unsigned long haps1[maxclasssize*maxfem+1], haps2[maxclasssize*maxfem+1]; 
unsigned long hind[maxclasssize*maxfem+1];
/* The arrays invgenenum1 and invgenenum2 store an effective inverse
to the genenumber mapping above. They convert an index into
genenumber into two haplotypes */
unsigned invgenenum1[maxfem], invgenenum2[maxfem]; /* Ramana */



 pathway thispath;
 boolean informative[maxped];
 boolean rare[maxfem], risk1[maxfem], risk2[maxfem];
 boolean riskmale[maxhap];
 approxrec *approxstruct;
 censorrec *censorstruct;
 long thisc;
 boolean malechild[maxchild];
 thisarray *thischild[maxchild];
 long nchild;
 long seghap1[maxfem], seghap2[maxfem];
 long segstart[maxfem];
 unsigned long probstart[maxfem], probend[maxfem];
 boolean nohom[maxlocus];
/* locusvalues *thislocus[maxlocus]; */
 long increment[maxlocus], order[maxlocus];
/*PEOPLE*/
/* thisperson *person[maxind + 1];
 thisperson *proband[maxped];
 thisperson *looppers[maxped][maxloop][2]; */
/*MUTATION */
 uchar muthap[maxhap];
 gennurec *gennustruct;
/*RECOMBINATION*/
 thetavalues *maletheta, *femaletheta;
/*FREQUENCIES */
 thisarray *hapfreq;
/*RISK*/
 char riskall;
/*OTHERS*/
 long risksys, mutsys, mlocus, lastpriv, i;
 long nuhap;
 long fgeno, mgeno;
 long nuped, totperson;
 double segscale, mutmale, mutfemale, like, alike, distratio;
 boolean interfer, disequi, sexlink, risk, sexdif, readfemale,
	        dolod, firstapprox, firsttime, lasttime;
 FILE *outfile, *ipedfile, *datafile, *stream, *speedfile;
/*ILINK*/
 FILE *final;
 boolean mapping;

/* Local variables for seg: */
/*

struct LOC_seg {
  struct LOC_likelihood *LINK;
  thisperson **p, **q, **r, *child, *father, *mother;
  long fseg, sseg, sstart, send, fstart, fend, nfirst, nsecond, firstseg,
       secondseg;
  double pf, ps;
  thetavalues *firstsex, *secondsex;
} ; */

/*Local variables for likelihood*/
/*
struct LOC_likelihood {
  long thisped;
  thisperson *proband;
  long loopgen[maxloop];
  double homo, hetero;
  long nuscale;
  thisarray *holdpoint[maxloop];
} ;
*/

extern void getprobtable();
extern void segup();
extern void segdown();
extern void segsexup();
extern void segsexdown();
extern void   performCheckpoint();
extern void  recoverCheckpoint ();
extern void getvect();
extern void recombination();
extern void getlocations();
extern void readspeed();
extern void invert();
extern void cleanup();
extern void initseg();
extern void exitseg();
extern void segsexctop();
extern void segsextop();
extern void segctop();
extern void segtop();
extern void segcapprox();
extern void msegsexdown();
extern void msegdown();
extern double mapfunction();
extern double getdist();
extern double invdist();
extern void setparam();
extern void inputdata();
extern void oldsegsexup();
extern void oldsegup();


